Require Import base ord data.fintype pts.category.

Section STLC.

Variable (Σ : Type). (* constant types *)
Variable (ext : Σ -> Type). (* elements of constant types *)

Inductive ty :=
| DefTy (* currently only defined a single type *)
| Arrow (A B : ty)
| Prod (σ : Σ) (B : ty).

Fixpoint pos (Γ : seq ty) (A : ty) : Type :=
  match Γ with
  | [::] => False
  | B :: Δ => (B = A) + pos Δ A
  end.

Unset Elimination Schemes.
Inductive tmn (Γ : seq ty) : ty -> Type :=
| var (A : ty) (i : pos Γ A) : tmn Γ A
| app (A B : ty) (s : tmn Γ (Arrow A B)) (t : tm Γ A) : tmn Γ B
| proj (σ : Σ) (B : ty) (s : tmn Γ (Prod σ B)) (v : ext σ) : tmn Γ B
with tm (Γ : seq ty) : ty -> Type :=
| lam (A B : ty) (s : tm (A :: Γ) B) : tm Γ (Arrow A B)
| abs (c : Σ) (B : ty) (b : ext c -> tm Γ B) : tm Γ (Prod c B)                                   
| neu (s : tmn Γ DefTy) : tm Γ DefTy.
Set Elimination Schemes.

Scheme tmn_ind := Induction for tmn Sort Prop
  with tm_ind := Induction for tm Sort Prop.
Combined Scheme t_ind from tmn_ind, tm_ind.

Definition env Γ (X : ty -> Type) :=
  forall A, pos Γ A -> X A.

Definition ren (Γ Δ : seq ty) :=
  env Γ (pos Δ).
  (*forall A, pos Γ A -> pos Δ A.*)

Definition scons {Γ A} {X : ty -> Type} (x : X A) (σ : env Γ X) : env (A :: Γ) X :=
  fun B (p : (A = B) + pos Γ B) =>
    match p with
    | inl E => match E with erefl => x end
    | inr p' => σ B p'
    end.

Definition upren {Γ Δ : seq ty} (ξ : ren Γ Δ) : forall {A}, ren (A :: Γ) (A :: Δ) :=
  fun A B (p : (A = B) + pos Γ B) =>
    match p with
    | inl E => inl E
    | inr p' => inr (ξ B p')
    end.

Fixpoint tmn_ren {Γ A} (s : tmn Γ A) : forall {Δ}, ren Γ Δ -> tmn Δ A :=
  match s with
  | var _ i => fun Δ ξ => var (ξ _ i)
  | app _ _ s t => fun Δ ξ => app (tmn_ren s ξ) (tm_ren t ξ)
  | proj _ _ s v => fun Δ ξ => proj (tmn_ren s ξ) v
  end
with tm_ren {Γ A} (s : tm Γ A) : forall {Δ}, ren Γ Δ -> tm Δ A :=
  match s with
  | lam _ _ s => fun Δ ξ => lam (tm_ren s (upren ξ))
  | abs _ _ b => fun Δ ξ => abs (fun v => tm_ren (b v) ξ)
  | neu s => fun Δ ξ => neu (tmn_ren s ξ)
  end.

Definition subst (Γ Δ : seq ty) :=
  forall A, pos Γ A -> tmn Δ A.

Definition bound {A B Γ} (E : A = B) : pos (A :: Γ) B :=
  inl E.

Definition shift {A Γ} : ren Γ (A :: Γ) :=
  fun B p => inr p.

Definition up {Γ Δ} (τ : subst Γ Δ) : forall {A}, subst (A :: Γ) (A :: Δ) :=
  fun A B (p : (A = B) + pos Γ B) =>
    match p with
    | inl E => @var (A :: Δ) B (bound E)
    | inr p' => tmn_ren (τ _ p') shift
    end.

Fixpoint tmn_inst {Γ A} (s : tmn Γ A) : forall {Δ}, subst Γ Δ -> tmn Δ A :=
  match s with
  | var _ i => fun Δ τ => τ _ i
  | app _ _ s t => fun Δ ξ => app (tmn_inst s ξ) (tm_inst t ξ)
  | proj _ _ s v => fun Δ ξ => proj (tmn_inst s ξ) v
  end
with tm_inst {Γ A} (s : tm Γ A) : forall {Δ}, subst Γ Δ -> tm Δ A :=
  match s with
  | lam _ _ s => fun Δ ξ => lam (tm_inst s (up ξ))
  | abs _ _ b => fun Δ ξ => abs (fun v => tm_inst (b v) ξ)
  | neu s => fun Δ ξ => neu (tmn_inst s ξ)
  end.

Definition idren {Γ} : ren Γ Γ :=
  fun _ => id.

Lemma upren_id {Γ A} : @upren Γ Γ idren A = idren.
Proof. fext=>/=B. by case. Qed.
 
Lemma tmn_tm_ren_id Γ :
  (forall A (s : tmn Γ A), tmn_ren s idren = s) /\
  (forall A (s : tm Γ A), tm_ren s idren = s).
Proof.
  move: Γ. apply t_ind => //=.
  - by move=> Γ A B s -> t ->.
  - by move=> Γ σ B s -> v.
  - move=> Γ A B s ih. by rewrite upren_id ih.
  - move=> Γ c B b ih. congr abs. fext=> v. by rewrite ih.
  - by move=> Γ s ->.
Qed.

Lemma tmn_ren_id Γ A (s : tmn Γ A) : tmn_ren s idren = s.
Proof. by case: (tmn_tm_ren_id Γ). Qed.

Lemma tm_ren_id Γ A (s : tm Γ A) : tm_ren s idren = s.
Proof. by case: (tmn_tm_ren_id Γ). Qed.

Coercion var : pos >-> tmn.

Lemma upren_up A Γ Δ (ξ : ren Γ Δ) : up ξ (A := A) = upren ξ.
Proof. fext=> /=B [E|p] //=. Qed.

Lemma tmn_tm_ren_inst Γ :
  (forall A (s : tmn Γ A) Δ (ξ : ren Γ Δ), tmn_ren s ξ = tmn_inst s ξ) /\
  (forall A (s : tm Γ A) Δ (ξ : ren Γ Δ), tm_ren s ξ = tm_inst s ξ).
Proof.
  move: Γ. apply: t_ind => //=.
  - move=> Γ A B s ihs t iht Δ ξ. by rewrite ihs iht.
  - move=> Γ σ B s ihs v Δ ξ. by rewrite ihs.
  - move=> Γ A B s ihs Δ ξ. by rewrite upren_up ihs.
  - move=> Γ c B b ih Δ ξ. congr abs; fext=> v. by rewrite ih.
  - move=> Γ s ih Δ ξ. by rewrite ih.
Qed.

Lemma tmn_ren_inst Γ Δ A (ξ : ren Γ Δ) (s : tmn Γ A) :
  tmn_ren s ξ = tmn_inst s ξ.
Proof. by case: (tmn_tm_ren_inst Γ). Qed.

Lemma tm_ren_inst Γ Δ A (ξ : ren Γ Δ) (s : tm Γ A) :
  tm_ren s ξ = tm_inst s ξ.
Proof. by case: (tmn_tm_ren_inst Γ). Qed.

Arguments var {Gamma} [A] p : rename.

Lemma tmn_inst_id Γ A (s : tmn Γ A) :
  tmn_inst s var = s.
Proof.
  by rewrite -tmn_ren_inst tmn_ren_id.
Qed.

Lemma tm_inst_id Γ A (s : tm Γ A) :
  tm_inst s var = s.
Proof.
  by rewrite -tm_ren_inst tm_ren_id.
Qed.

Lemma tmn_tm_ren_inst_comp Γ :
  (forall A (s : tmn Γ A) Δ Δ' (ξ : ren Γ Δ) (τ : subst Δ Δ'),
      tmn_inst (tmn_ren s ξ) τ = tmn_inst s (fun B p => τ B (ξ B p))) /\
  (forall A (s : tm Γ A) Δ Δ' (ξ : ren Γ Δ) (τ : subst Δ Δ'),
      tm_inst (tm_ren s ξ) τ = tm_inst s (fun B p => τ B (ξ B p))).
Proof.
  move: Γ. apply: t_ind => //=.
  - move=> Γ A B s ihs t iht Δ Δ' ξ τ. by rewrite ihs iht.
  - move=> Γ σ B s ihs v Δ Δ' ξ τ. by rewrite ihs.
  - move=> Γ A B s ihs Δ Δ' ξ τ. rewrite ihs. do 2 f_equal.
      by fext=>/= C [E|p].
  - move=> Γ c B b ih Δ Δ' ξ τ. congr abs. fext=> v. by rewrite ih.
  - move=> Γ s ihs Δ Δ' ξ τ. by rewrite ihs.
Qed.

Lemma tmn_ren_inst_comp Γ Δ Δ' A (ξ : ren Γ Δ) (τ : subst Δ Δ') (s : tmn Γ A) :
  tmn_inst (tmn_ren s ξ) τ = tmn_inst s (fun B p => τ B (ξ B p)).
Proof. by case: (tmn_tm_ren_inst_comp Γ). Qed.

Lemma tm_ren_inst_comp Γ Δ Δ' A (ξ : ren Γ Δ) (τ : subst Δ Δ') (s : tm Γ A) :
  tm_inst (tm_ren s ξ) τ = tm_inst s (fun B p => τ B (ξ B p)).
Proof. by case: (tmn_tm_ren_inst_comp Γ). Qed.

Lemma tmn_ren_comp Γ Δ Δ' A (ξ : ren Γ Δ) (ξ' : ren Δ Δ') (s : tmn Γ A) :
  tmn_ren (tmn_ren s ξ) ξ' = tmn_ren s (fun B p => ξ' B (ξ B p)).
Proof.
  by rewrite tmn_ren_inst tmn_ren_inst_comp -tmn_ren_inst.
Qed.

Lemma tm_ren_comp Γ Δ Δ' A (ξ : ren Γ Δ) (ξ' : ren Δ Δ') (s : tm Γ A) :
  tm_ren (tm_ren s ξ) ξ' = tm_ren s (fun B p => ξ' B (ξ B p)).
Proof.
  by rewrite tm_ren_inst tm_ren_inst_comp -tm_ren_inst.
Qed.

Lemma tmn_tm_inst_ren_comp Γ :
  (forall A (s : tmn Γ A) Δ Δ' (τ : subst Γ Δ) (ξ : ren Δ Δ'),
      tmn_ren (tmn_inst s τ) ξ = tmn_inst s (fun B p => tmn_ren (τ B p) ξ)) /\
  (forall A (s : tm Γ A) Δ Δ' (τ : subst Γ Δ) (ξ : ren Δ Δ'),
      tm_ren (tm_inst s τ) ξ = tm_inst s (fun B p => tmn_ren (τ B p) ξ)).
Proof.
  move: Γ. apply: t_ind => //=.
  - move=> Γ A B s ihs t iht Δ Δ' τ ξ. by rewrite ihs iht.
  - move=> Γ σ B s ihs v Δ Δ' τ ξ. by rewrite ihs.
  - move=> Γ A B s ihs Δ Δ' τ ξ. rewrite ihs. do 2 f_equal.
      fext=>/= C [E|p] //=. by rewrite !tmn_ren_comp.
  - move=> Γ c B b ih Δ Δ' τ ξ. congr abs. fext=> v. by rewrite ih.
  - move=> Γ s ihs Δ Δ' τ ξ. by rewrite ihs.
Qed.

Lemma tmn_inst_ren_comp Γ Δ Δ' A (τ : subst Γ Δ) (ξ : ren Δ Δ') (s : tmn Γ A) :
  tmn_ren (tmn_inst s τ) ξ = tmn_inst s (fun B p => tmn_ren (τ B p) ξ).
Proof. by case: (tmn_tm_inst_ren_comp Γ). Qed.

Lemma tm_inst_ren_comp Γ Δ Δ' A (τ : subst Γ Δ) (ξ : ren Δ Δ') (s : tm Γ A) :
  tm_ren (tm_inst s τ) ξ = tm_inst s (fun B p => tmn_ren (τ B p) ξ).
Proof. by case: (tmn_tm_inst_ren_comp Γ). Qed.

Lemma tmn_tm_inst_comp Γ :
  (forall A (s : tmn Γ A) Δ Δ' (τ : subst Γ Δ) (τ' : subst Δ Δ'),
      tmn_inst (tmn_inst s τ) τ' = tmn_inst s (fun B p => tmn_inst (τ B p) τ')) /\
  (forall A (s : tm Γ A) Δ Δ' (τ : subst Γ Δ) (τ' : subst Δ Δ'),
      tm_inst (tm_inst s τ) τ' = tm_inst s (fun B p => tmn_inst (τ B p) τ')).
Proof.
  move: Γ. apply: t_ind => //=.
  - move=> Γ A B s ihs t iht Δ Δ' τ τ'. by rewrite ihs iht.
  - move=> Γ σ B s ihs v Δ Δ' τ τ'. by rewrite ihs.
  - move=> Γ A B s ihs Δ Δ' τ τ'. rewrite ihs. do 2 f_equal.
      fext=>/= C [E|p] //=. by rewrite tmn_inst_ren_comp tmn_ren_inst_comp.
  - move=> Γ c B b ih Δ Δ' τ τ'. congr abs. fext=> v. by rewrite ih.
  - move=> Γ s ihs Δ Δ' τ τ'. by rewrite ihs.
Qed.

Lemma tmn_inst_comp Γ Δ Δ' A (τ : subst Γ Δ) (τ' : subst Δ Δ') (s : tmn Γ A) :
  tmn_inst (tmn_inst s τ) τ' = tmn_inst s (fun B p => tmn_inst (τ B p) τ').
Proof. by case: (tmn_tm_inst_comp Γ). Qed.

Lemma tm_inst_comp Γ Δ Δ' A (τ : subst Γ Δ) (τ' : subst Δ Δ') (s : tm Γ A) :
  tm_inst (tm_inst s τ) τ' = tm_inst s (fun B p => tmn_inst (τ B p) τ').
Proof. by case: (tmn_tm_inst_comp Γ). Qed.

(** ** Normalization by Evaluation *)

(*Structure prshf : Type :=
  { prshf_domain :> seq ty -> Type
  ; map : forall {Γ Δ} (ξ : ren Γ Δ), prshf_domain Γ -> prshf_domain Δ
  }.

Definition tmn_prshf A : prshf :=
  {| prshf_domain := fun Γ => tmn Γ A
   ; map := fun Γ Δ ξ u => tmn_ren u ξ
   |}.

Structure prshf_exp (A B : prshf) (Γ : seq ty) : Type :=
  { prshf_exp_domain :> forall Δ, ren Γ Δ -> A Δ -> B Δ
  ; prshf_exp_nat : forall Δ Θ (ξ : ren Γ Δ) (ζ : ren Δ Θ) (a : A Δ),
      map ζ (prshf_exp_domain ξ a) = prshf_exp_domain (fun B i => ζ B (ξ B i)) (map ζ a)
  }.

Definition exp_prshf (A B : prshf) : prshf :=
  {| prshf_domain := prshf_exp A B
   ; map := fun Γ Δ ξ u => *)

Fixpoint D Γ A : Type :=
  match A with
  | DefTy => tmn Γ DefTy
  | Arrow A B => forall Δ, ren Γ Δ -> D Δ A -> D Δ B
  | Prod σ A => ext σ -> D Γ A
  end.

Definition dsubst Γ Δ : Type :=
  forall A, pos Γ A -> D Δ A.

Definition d_ren A :
  forall {Γ Δ}, ren Γ Δ -> D Γ A -> D Δ A.
Proof.
  induction A; cbn.
  - intros Γ Δ ξ u. exact (tmn_ren u ξ).
  - intros Γ Δ ξ f Θ ζ a. exact (f Θ (fun B i => ζ B (ξ B i)) a).
  - intros Γ Δ ξ f i. exact (IHA Γ Δ ξ (f i)).
Defined.

Definition dshift Γ Δ A (σ : dsubst Γ Δ) : dsubst Γ (A :: Δ) :=
  fun B i => d_ren shift (σ B i).

Fixpoint d_inst {Γ Δ A} (s : tm Γ A) (σ : dsubst Γ Δ) : D Δ A :=
  match s in tm _ A return D Δ A with
  | neu u => dn_inst u σ
  | lam _ _ s => fun Δ ξ a => d_inst s (scons a (fun B i => d_ren ξ (σ B i)))
  | abs _ _ f => fun i => d_inst (f i) σ
  end
with dn_inst {Γ Δ A} (u : tmn Γ A) (σ : dsubst Γ Δ) : D Δ A :=
  match u in tmn _ A return D Δ A with
  | var _ i => σ _ i
  | app _ _ u s => (dn_inst u σ) Δ (fun _ => id) (d_inst s σ)
  | proj _ _ u i => dn_inst u σ i
  end.

Definition reflect_reify (A : ty) :
  forall Γ, (tmn Γ A -> D Γ A) * (D Γ A -> tm Γ A).
Proof.
  induction A; cbn; intro Γ.
  - exact (id, @neu Γ).
  - split. intros u Δ ξ a. refine ((IHA2 Δ).1 (app (tmn_ren u ξ) ((IHA1 Δ).2 a))).
    intro f. refine (lam _). apply ((IHA2 _).2). exact (f (A1 :: Γ) shift ((IHA1 (A1 :: Γ)).1 (var (bound erefl)))).
  - split. intros u i. exact ((IHA Γ).1 (proj u i)).
    intros f. apply abs. intros i. apply (IHA Γ).2. exact (f i).
Defined.

Definition reflect {Γ A} (u : tmn Γ A) : D Γ A :=
  (reflect_reify A Γ).1 u.

Definition reify {Γ A} (a : D Γ A) : tm Γ A :=
  (reflect_reify A Γ).2 a.

Definition embed {Γ A} (s : tm Γ A) :
  D Γ A.
Proof.
  revert Γ A s. fix embed 3. intros Γ A s. destruct s; cbn.
  - intros Δ ξ a. refine (d_inst s (scons a _)). intros C i. exact (reflect (ξ C i)).
  - intros i. apply embed. exact (b i).
  - exact s.
Defined.

Definition hinst {Γ Δ A} (s : tm Γ A) (σ : env Γ (tm Δ)) : tm Δ A :=
  reify (d_inst s (fun B i => embed (σ B i))).

Definition App {Γ A B} (s : tm Γ (Arrow A B)) (t : tm Γ A) : tm Γ B :=
  reify (embed s Γ (fun _ => id) (embed t)).


(** ** Hereditary Substitutions *)

Fixpoint tmr Γ A : Type :=
  match A with
  | DefTy => tmn Γ A
  | Arrow A B => tmr (A :: Γ) B
  | Prod σ B => ext σ -> tmr Γ B
  end.

Fixpoint tm_to_tmr {Γ A} (s : tm Γ A) : tmr Γ A :=
  match s in tm _ A return tmr Γ A with
  | neu t => t
  | lam _ _ b => tm_to_tmr b
  | abs _ _ b => fun i => tm_to_tmr (b i)
  end.

Fixpoint tmr_to_tm {Γ A} : tmr Γ A -> tm Γ A :=
  match A as A return tmr Γ A -> tm Γ A with
  | DefTy => @neu _
  | Arrow A B => fun b => lam (@tmr_to_tm (A :: Γ) B b)
  | Prod σ B => fun b => abs (fun i => tmr_to_tm (b i))
  end.

Lemma tm_to_tmr_inv Γ A (s : tm Γ A) :
  tmr_to_tm (tm_to_tmr s) = s.
Proof.
  move: Γ A s. apply: (tm_ind (P := fun _ _ _ => True)) => //=.
  - by move=> Γ A B s ->.
  - move=> Γ c B b ih. congr abs; fext=> i. exact: ih.
Qed.

Lemma tmr_to_tm_inv Γ A (s : tmr Γ A) :
  tm_to_tmr (tmr_to_tm s) = s.
Proof.
  elim: A Γ s => //= σ B ih Γ b. fext=> i. exact: ih.
Qed.

Definition hsubst (Γ Δ : seq ty) := forall A, pos Γ A -> tmr Δ A.

Program Fixpoint tneu {Γ A} {struct A} : tmn Γ A -> tmr Γ A :=
  match A as A return tmn Γ A -> tmr Γ A with
  | DefTy => id
  | Arrow A B =>
    fun s =>
      @tneu (A :: Γ) B (app (tmn_ren s shift) (tmr_to_tm (@tneu (A :: Γ) A (var (inl erefl))))) 
  | Prod σ B => fun s i => @tneu Γ B (proj s i)
  end.

Definition tvar {Γ A} (i : pos Γ A) : tmr Γ A :=
  tneu (var i).

Definition hsubst_up {Γ Δ A} (τ : hsubst Γ Δ) : hsubst (A :: Γ) (A :: Δ).
Proof.
  intros B [E|i]. apply: tvar. cbn. exact (inl E).
  specialize (τ B i). apply: tm_to_tmr. refine (tm_ren _ shift). apply: tmr_to_tm. exact: τ.
Defined.

Section HereditarySubstitution.
Variable tmn_hinst_defty : forall Γ Δ (s : tmn Γ DefTy) (τ : hsubst Γ Δ), tmn Δ DefTy.

Fixpoint tmr_hinst {A} : forall Γ Δ, tmr Γ A -> hsubst Γ Δ -> tmr Δ A :=
  match A as A return forall Γ Δ, tmr Γ A -> hsubst Γ Δ -> tmr Δ A with
  | DefTy => tmn_hinst_defty
  | Arrow A B =>
    fun Γ Δ (s : tmr (A :: Γ) B) τ => tmr_hinst s (hsubst_up τ)
  | Prod σ B =>
    fun Γ Δ (b : ext σ -> tmr Γ B) τ i => tmr_hinst (b i) τ
  end.

Definition tmr_app {Γ A B} (s : tmr Γ (Arrow A B)) (t : tmr Γ A) : tmr Γ B :=
  @tmr_hinst B (A :: Γ) Γ s (fun T p => match p with inl E => match E with erefl => t end | inr i => tvar i end).

Fixpoint tmn_hinst {Γ A} (s : tmn Γ A) : forall {Δ}, hsubst Γ Δ -> tmr Δ A :=
  match s with
  | var _ i => fun Δ τ => τ _ i
  | app _ _ s t => fun Δ τ => tmr_app (tmn_hinst s τ) (tmr_hinst (tm_to_tmr t) τ)
  | proj _ _ s v => fun Δ τ => tmn_hinst s τ v
  end.
End HereditarySubstitution.

Section PresheafInterpretation.

Fixpoint L (A : ty) : seq ty -> Type :=
  match A with
  | DefTy => fun Γ => tmn Γ DefTy
  | Arrow A B => fun Γ => forall Δ, ren Γ Δ -> L A Δ -> L B Δ
  | Prod σ B => fun Γ => ext σ -> L B Γ
  end.

Fixpoint G (Γ : seq ty) : seq ty -> Type :=
  match Γ with
  | [::] => fun _ => unit
  | A :: Γ => fun Δ => (L A Δ * G Γ Δ)%type
  end.

Definition G_pos Γ : forall A Δ, pos Γ A -> G Γ Δ -> L A Δ.
Proof.
  induction Γ => //= A Δ [E|i]//=. destruct E. apply fst. intros p. apply (IHΓ _ _ i (snd p)).
Defined.

Definition L_ren A : forall {Γ Δ}, ren Γ Δ -> L A Γ -> L A Δ.
  induction A => //=.
  - move=> Γ Δ ξ s. exact: tmn_ren ξ.
  - move=> Γ Δ ξ ih Θ ξ' s. apply: ih. move=> B i. apply: ξ'. exact: ξ. exact: s.
  - move=> Γ Δ ξ b i. exact: IHA (b i).
Defined.

Definition G_ren Γ : forall {Δ Θ}, ren Δ Θ -> G Γ Δ -> G Γ Θ.
  induction Γ => //= Δ Θ ξ [p q]. split. exact: L_ren p. exact: IHΓ q.
Defined.

Fixpoint L_inst {Γ A} (s : tmn Γ A) : forall Δ, G Γ Δ -> L A Δ :=
  match s in tmn _ A return forall Δ, G Γ Δ -> L A Δ with
  | var A i => fun Δ ρ => G_pos i ρ
  | app A B s t => fun Δ ρ => @L_inst Γ (Arrow A B) s Δ ρ Δ idren (@T_inst Γ A t Δ ρ)
  | proj σ B b i => fun Δ ρ => @L_inst Γ (Prod σ B) b Δ ρ i
  end
with T_inst {Γ A} (s : tm Γ A) : forall Δ, G Γ Δ -> L A Δ :=
  match s in tm _ A return forall Δ, G Γ Δ -> L A Δ with
  | neu t => L_inst t
  | lam A B b => fun Δ ρ Θ ξ a => @T_inst (A :: Γ) B b Θ (a, G_ren ξ ρ)
  | abs σ B b => fun Δ ρ i => @T_inst Γ B (b i) Δ ρ
  end.

Scheme tmn_rect := Induction for tmn Sort Type.

Definition quote_unquote A :
  (forall Γ, tmn Γ A -> L A Γ) * (forall Γ, L A Γ -> tm Γ A).
Proof.
  induction A as [|A [ihA1 ihA2] B [ihB1 ihB2]|σ B [ih1 ih2]] => /=.
  - intuition. exact: neu.
  - split.
    + move=> Γ s Δ ξ a. specialize (ihA2 _ a). apply: ihB1. apply: app ihA2. exact: (tmn_ren s ξ).
    + move=> Γ H. constructor. apply: ihB2. apply: H. exact shift. apply: ihA1. apply: var. exact: inl.
  - split.
    + move=> Γ s i. apply: ih1. exact: proj s i.
    + move=> Γ H. constructor=> i. apply: ih2. exact: H.
Defined.

Definition G_id Γ : G Γ Γ.
  induction Γ => /=. exact tt. split. apply: (quote_unquote a).1. apply: var. exact: inl. exact: G_ren shift IHΓ.
Defined.

Definition quote_tm Γ A (s : tm Γ A) : L A Γ.
  apply: (T_inst s). exact: G_id.
Defined.

Definition G_subst Γ : forall Δ, (forall A, pos Γ A -> tm Δ A) -> G Γ Δ.
  induction Γ => //=Δ τ. split. apply: quote_tm. apply: τ. by left.
  apply: IHΓ => A i. apply: τ. by right.
Defined.

Definition nat_ty : ty := Arrow (Arrow DefTy DefTy) (Arrow DefTy DefTy).

Definition zero : tm [::] nat_ty. repeat constructor. Defined.

Definition succ : tm [::] (Arrow nat_ty nat_ty).
  constructor. constructor. constructor. constructor.
  eapply app. constructor. right. by left.
  constructor. eapply app. eapply app. constructor. right. right. by left.
  constructor. constructor. eapply app. constructor. right. right. by left.
  constructor. constructor. by left.
  constructor. constructor. by left.
Defined.

Definition tapp Γ A B (s : tm Γ (Arrow A B)) (t : tm Γ A) : tm Γ B :=
  match s with
  | lam _ _ b => 

Definition four : tm [::] nat_ty

Definition mul : tm [::] (Arrow nat_ty (Arrow nat_ty nat_ty)).
  constructor. constructor. constructor. constructor. constructor.
  eapply app. eapply app. apply: var. right. right. right. left. reflexivity.
  constructor. constructor. eapply app. apply: var. right. right. by left. constructor. apply: var. by left.
  constructor. apply: var. by left.
Defined.

Definition hinst {Γ A} (s : tm Γ A) : forall Δ, (forall B, pos Γ B -> tm Δ B) -> tm Δ A.
intros Δ H. apply: (quote_unquote A).2. apply: (T_inst s).

Definition tmn_to_L {Γ A} (i : tmn Γ A) : L A Γ.
Proof.
  elim: A Γ i => //=.
  move=> A ihA B ihB Γ i Δ ξ a.
  elim: Γ A i => //A Γ ih B [E|/ih]. destruct E. apply: ih.
  elim: s => {Γ A} /=.

Definition tmr_to_L {Γ A} (s : tmr Γ A) : L A Γ.
  elim: A Γ s => [|A ihA B ihB|σ B ih] Γ /= s.
  - exact: s.
  - move=> Δ ξ a.

Definition hsusbt_to_G {Γ Δ} (τ : hsubst Γ Δ) : G Γ Δ.
Proof.
  elim: Γ Δ τ => //= A Γ ih Δ τ. specialize (ih Δ (fun _ i => τ _ (inr i))).
  split=>//. move: (τ A (inl erefl)).
Section LogicalInterpretation.
Variable X : Type.

Fixpoint L (A : ty) : Type :=
  match A with
  | DefTy => X
  | Arrow A B => L A -> L B
  | Prod σ B => ext σ -> L B
  end.

Fixpoint G (Γ : seq ty) : Type :=
  match Γ with
  | [::] => True
  | A :: Γ => (L A * G Γ)%type
  end.

Definition G_pos Γ : forall A, pos Γ A -> G Γ -> L A.
Proof.
  induction Γ => //= A [E|i]//=. destruct E. apply fst. intros p. apply (IHΓ _ i (snd p)).
Defined.

Fixpoint L_inst {Γ A} (s : tmn Γ A) : G Γ -> L A :=
  match s in tmn _ A return G Γ -> L A with
  | var A i => fun ρ => G_pos i ρ
  | app A B s t => fun ρ => @L_inst Γ (Arrow A B) s ρ (@T_inst Γ A t ρ)
  | proj σ B b i => fun ρ => @L_inst Γ (Prod σ B) b ρ i
  end
with T_inst {Γ A} (s : tm Γ A) : G Γ -> L A :=
  match s in tm _ A return G Γ -> L A with
  | neu t => L_inst t
  | lam A B b => fun ρ a => @T_inst (A :: Γ) B b (a, ρ)
  | abs σ B b => fun ρ i => @T_inst Γ B (b i) ρ
  end.

End LogicalInterpretation.

Scheme tmn_rect := Induction for tmn Sort Type.

Definition hcons {Γ Δ A} (s : tmr Δ A) (τ : hsubst Γ Δ) : hsubst (A :: Γ) Δ :=
  fun B p => match p with
  | inl E => match E with erefl => s end
  | inr i => τ B i
  end.

Definition tmn_G_inst {Γ Δ} (s : tmn Γ DefTy) : G (tmn Δ DefTy) Γ -> tmn Δ DefTy :=
  L_inst s.

Definition tmr_to_L {Γ Δ A} (s : tmr Δ A) :
  G (tmn Δ DefTy) Γ -> L (tmn Δ DefTy) A.
Proof.
  elim: A Γ Δ s => /=.
  - move=> A ihA B ihB Γ Δ s ρ a.
    specialize (ihB (A :: Γ) _ s). move: ihB => /=.

Definition hsusbt_to_G {Γ Δ} (τ : hsubst Γ Δ) : G (tmn Δ DefTy) Γ.
Proof.
  elim: Γ Δ τ => //A Γ ih Δ τ. specialize (ih Δ (fun _ i => τ _ (inr i))).
  cbn. split=>//. move: (τ A (inl erefl)).

  Definition tmr_to_L {Γ Δ A} (s : tmr Δ A) (τ : hsubst Γ Δ) : G (tmn Δ DefTy) Γ -> L (tmn Δ DefTy) A.
Definition lvar {Γ Δ A} (i : pos Γ A) (τ : hsubst Γ Δ) : L (tmr Δ DefTy) A.
Proof.
  elim: A Γ Δ i τ => /= [|A ihA B ihB|σ B ih] Γ Δ i τ.
  - exact: (τ _ i).
  - move=> a. specialize (ihB _ _ _ (hcons eapply ihB.

Definition lvar {Γ A} (i : tmn Γ A) : L (tmn Γ DefTy) A.
  move: Γ A i. apply: tmn_rect.
  - move=> Γ A i /=.
  apply (@tmn_rect (fun Γ A _ => L (tmn Γ DefTy) A) (fun _ _ _ => True)). 
  elim: A Γ i => /=.
  - move=> Γ i. exact: i.
  - move=> A ihA B ihB Γ s a. exact: ihB.
  - move=> σ B ih Γ b i.
elim: Γ A i => //=B Γ ih A [E|i].

Definition embed {Γ A} (s : tmn Γ A) : 

Fixpoint hinstT Γ (A : ty) X : Type :=
  match A with
  | DefTy => X
  | Arrow A B => tmr Γ A -> hinstT Γ B X
  | Prod σ B => hinstT Γ B X
  end.

Program Fixpoint tmn_hinst_base {Γ A} (s : tmn Γ A) :
  forall Δ, hsubst Γ Δ -> hinstT Γ A (tmn Δ DefTy) :=
  match s as s in tmn _ A return forall Δ, hsubst Γ Δ -> hinstT Γ A (tmn Δ DefTy) with
  | var _ i => _
  | app _ _ s t => fun Δ τ => @tmn_hinst_base _ _ s Δ τ _
  | proj _ _ s v => _
  end.
Next Obligation.
  cbn.
Admitted.
Next Obligation.
  cbn.
cbn.

Fixpoint tmn_hinst_defty {Γ} (s : tmn Γ DefTy) : forall Δ, hsubst Γ Δ -> tmn Δ DefTy :=
  match s with
  | var _ i => fun Δ τ => τ _ i
  | app _ _ s t => tmn_hinst s



(*** Presheaf Interpretation ***)
(*Set Universe Polymorphism.
Unset Universe Minimization ToSet.*)

(*Require Import pts.category.*)

(** * The category of contexts *)

Definition ctx : Type := seq ty.
Definition cmorph (m n : ctx) : Type := subst n m.
Definition ctxid (m : ctx) : cmorph m m := var.
Definition ccomp m n k (f : cmorph m n) (g : cmorph n k) : cmorph m k :=
  fun A p => tmn_inst (g A p) f.

Lemma ccomp_idL m n (f : cmorph m n) : ccomp (@ctxid _) f = f.
Proof. fext=> A p /=. by rewrite/ccomp tmn_inst_id. Qed.

Lemma ccomp_idR m n (f : cmorph m n) : ccomp f (@ctxid _) = f.
Proof. by []. Qed.

Lemma ccomp_assoc m n k l (f : cmorph m n) (g : cmorph n k) (h : cmorph k l) :
  ccomp (ccomp f g) h = ccomp f (ccomp g h).
Proof. fext=> A p /=. by rewrite/ccomp tmn_inst_comp. Qed.

Canonical ctx_catMixin := @CatMixin ctx cmorph ctxid ccomp
  ccomp_idL ccomp_idR ccomp_assoc.
Canonical CTX := Eval hnf in Cat ctx cmorph ctx_catMixin.

(** *** Products in contexts *)

Definition ctx_prod : ctx -> ctx -> ctx := seq.cat.
Definition ctx_terminal : ctx := [::].

Fixpoint ctx_prod_split Γ : forall Δ A, pos (ctx_prod Γ Δ) A -> pos Γ A + pos Δ A :=
  match Γ as Γ return forall Δ A, pos (ctx_prod Γ Δ) A -> pos Γ A + pos Δ A with
  | [::] => fun Δ A p => inr p
  | B :: Γ' => fun Δ A ep =>
    match ep with
    | inl E => inl (inl E)
    | inr p =>
      match @ctx_prod_split Γ' Δ A p with
      | inl l => inl (inr l)
      | inr r => inr r
      end
    end
  end.

Definition ctx_inl Γ Δ A (p : pos Γ A) : pos (ctx_prod Γ Δ) A.
Proof.
  induction Γ; cbn in *. destruct p. destruct p as [E|p].
  apply inl. exact E. apply inr. apply IHΓ. apply p.
Defined.

Definition ctx_inr Γ Δ A (p : pos Δ A) : pos (ctx_prod Γ Δ) A.
Proof.
  induction Γ; cbn in *. exact p. apply shift. exact IHΓ.
Defined.

Lemma ctx_inl_split Γ Δ A (p : pos Γ A) :
  ctx_prod_split (@ctx_inl Γ Δ A p) = inl p.
Proof.
  induction Γ; cbn in *. destruct p. destruct p as [E|p]; cbn=>//.
    by rewrite (IHΓ p).
Qed.

Lemma ctx_inr_split Γ Δ A (p : pos Δ A) :
  ctx_prod_split (@ctx_inr Γ Δ A p) = inr p.
Proof.
  induction Γ; cbn in * => //. by rewrite IHΓ.
Qed.

Lemma ctx_split_eta Γ Δ A (p : pos (ctx_prod Γ Δ) A) :
  match ctx_prod_split p with
  | inl j => ctx_inl Δ j
  | inr j => ctx_inr Γ j
  end = p.
Proof.
  induction Γ; cbn in *=>//. destruct p as [E|p]; cbn=>//.
  move: (IHΓ p). case: (ctx_prod_split p) => k eqn;
    by rewrite -eqn.
Qed.

Definition ctx_pi1 (Γ Δ : ctx) : hom (ctx_prod Γ Δ) Γ.
  move=> A p. exact: ctx_inl.
Defined.

Definition ctx_pi2 (Γ Δ : ctx) : hom (ctx_prod Γ Δ) Δ.
  move=> A p. exact: ctx_inr.
Defined.

Definition ctx_pair (Γ Δ Θ : ctx) (f : hom Γ Δ) (g : hom Γ Θ) : hom Γ (ctx_prod Δ Θ).
  move=> A p. destruct (ctx_prod_split p) as [j|j]. exact (f _ j). exact (g _ j).
Defined.

Definition ctx_pair_pi1 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi1 _ _ = f.
Proof.
  fext=> A p /=. cbn. unfold ctx_pair. by rewrite ctx_inl_split.
Qed.

Definition ctx_pair_pi2 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi2 _ _ = g.
Proof.
  fext=> A p /=. cbn. unfold ctx_pair. by rewrite ctx_inr_split.
Qed.

Definition ctx_pair_eta (m n k : ctx) (f : hom m (ctx_prod n k)) :
  ctx_pair (f ∘ @ctx_pi1 _ _) (f ∘ @ctx_pi2 _ _) = f.
Proof.
  fext=> A p /=. cbn. unfold ctx_pair. rewrite -{2}[p]ctx_split_eta.
    by case: (ctx_prod_split p).
Qed.

Definition ctx_bang (n : ctx) : hom n ctx_terminal := fun _ p => match p with end.

Lemma ctx_terminal_eta (n : ctx) (f : hom n ctx_terminal) :
  f = ctx_bang n.
Proof. by fext. Qed.

Canonical ctx_cartcatMixin :=
  @CartCat.Mixin ctx cmorph ctx_catMixin
    ctx_terminal ctx_bang ctx_prod ctx_pi1 ctx_pi2 ctx_pair
    ctx_pair_pi1 ctx_pair_pi2 ctx_pair_eta
    ctx_terminal_eta.
Canonical ctx_cartcat :=
  Eval hnf in CartCat.pack ctx_cartcatMixin.

(** * Presheaf of terms *)

Definition TM@{i j k} (A : ty) : presheaf@{i j k} ctx_cartcat :=
  mk_presheaf (fun (Γ : ctx_cartcat) => tm Γ A) (fun Γ Δ f s => tm_inst s f) (fun Γ s => tm_inst_id s) (fun Γ Δ Δ' f g s => tm_inst_comp g f s).

Definition TMN@{i j k} (A : ty) : presheaf@{i j k} ctx_cartcat :=
  mk_presheaf (fun (Γ : ctx_cartcat) => tmn Γ A) (fun Γ Δ f s => tmn_inst s f) (fun Γ s => tmn_inst_id s) (fun Γ Δ Δ' f g s => tmn_inst_comp g f s).

(*Program Definition Tmn'@{i j k} (A : ty) : presheaf@{i j k} ctx_cartcat :=
  mk_presheaf (fun Γ => tmn Γ A) (fun Γ Δ f s => tmn_inst s f) _ _.
Next Obligation.
  exact: tmn_inst_id.
Qed.
Next Obligation.
  unfold comp => /=. unfold ccomp => /=.
  exact: tmn_inst_comp.
(* Error:
Anomaly
"File "engine/uState.ml", line 353, characters 28-34: Assertion failed."
Please report at http://coq.inria.fr/bugs/. *)
Qed.*)

(** *** Families in presheafs *)

Section PresheafFam.

Context {C : cat} (A : Type).

Program Definition Fam (p : presheaf C) : presheaf C :=
  mk_presheaf (fun c => A -> p c) (fun _ _ f p x => map f (p x)) _ _.
Next Obligation.
  fext=> i. by csimpl.
Qed.
Next Obligation.
  fext=> i. by csimpl.
Qed.

Program Definition Fam_proj (p : presheaf C) (x : A) : pnat (Fam p) p :=
  mk_pnat (fun c f => f x) (fun _ _ _ _ => erefl).

Program Definition Fam_abs (p q : presheaf C) (eta : A -> pnat p q) : pnat p (Fam q) :=
  mk_pnat (fun c pc x => eta x c pc) _.
Next Obligation.
  fext=> i /=. exact: natP.
Qed.

End PresheafFam.

(** *** Term model *)

Program Definition APP (A B : ty) : pnat (PProd (TMN (Arrow A B)) (TM A)) (TMN B) :=
  mk_pnat (fun n p => app p.1 p.2) (fun _ _ _ _ => erefl).

Program Definition LAM (A B : ty) : pnat (PExp (TM B) (TMN A)) (TM (Arrow A B)) :=
  mk_pnat (fun Γ F => lam (F (A :: Γ) (fun (C : ty) (p : pos Γ C) => shift p) (var (inl (erefl A))))) _.
Next Obligation.
  move: (svalP x) => /=->. cbn. repeat f_equal. fext=> C p.
  cbn. unfold ccomp. by rewrite tmn_ren_inst.
Qed.

(* note: the way things are set up, NEU is an equivalence *)
Program Definition NEU (A : ty) : pnat (TMN DefTy) (TM DefTy) :=
  mk_pnat (fun Γ s => neu s) (fun _ _ _ _ => erefl).

Program Definition PROJ (σ : Σ) (B : ty) : pnat (TMN (Prod σ B)) (Fam (ext σ) (TMN B)) :=
  mk_pnat (fun Γ s x => STLC.proj s x) (fun _ _ _ _ => erefl).

Program Definition ABS (σ : Σ) (B : ty) : pnat (Fam (ext σ) (TM B)) (TM (Prod σ B)) :=
  mk_pnat (fun Γ f => abs f) (fun _ _ _ _ => erefl).

(*** Presheaf Interpretation ***)

Section Interpretation.

Variable (C : cat).
Variable (def : presheaf C).

Fixpoint Ty (A : ty) : presheaf C :=
  match A with
  | DefTy => def
  | Arrow A B => PExp (Ty B) (Ty A)
  | Prod σ B => Fam (ext σ) (Ty B)
  end.

Fixpoint Ctx (Γ : seq ty) : presheaf C :=
  match Γ with
  | [::] => cunit
  | A :: Δ => Ctx Δ × Ty A
  end.

Definition Pos (Γ : seq ty) (A : ty) (p : pos Γ A) : hom (Ctx Γ) (Ty A).
  induction Γ. destruct p. cbn in p. destruct p as [E|p].
  - exact (match E with erefl => π₂ end).
  - exact (π₁ ∘ (IHΓ p)).
Defined.

Fixpoint Tmn {Γ : seq ty} {A : ty} (s : tmn Γ A) : pnat (Ctx Γ) (Ty A) :=
  match s in tmn _ A return pnat (Ctx Γ) (Ty A) with
  | var _ i => Pos i
  | app A B s t => ⟨ Tmn s, Tm t ⟩ ∘ nat_apply (Ty B) (Ty A)
  | STLC.proj σ B s v => Tmn s ∘ Fam_proj (Ty B) v
  end
with Tm {Γ : seq ty} {A : ty} (s : tm Γ A) : pnat (Ctx Γ) (Ty A) :=
  match s in tm _ A return pnat (Ctx Γ) (Ty A) with
  | lam A B s => nat_lam (Tm s)
  | abs c B b => Fam_abs (fun v => Tm (b v))
  | neu s => Tmn s
  end.

Fixpoint Ren {Γ : seq ty} : forall {Δ : seq ty}, ren Γ Δ -> pnat (Ctx Δ) (Ctx Γ) :=
  match Γ as Γ return forall {Δ : seq ty}, ren Γ Δ -> pnat (Ctx Δ) (Ctx Γ) with
  | [::] => fun _ _ => !
  | A :: Γ => fun Δ ξ => ⟨ Ren (fun B i => ξ B (inr i)), Pos (ξ A (inl erefl)) ⟩
  end.

Fixpoint Subst {Γ : seq ty} : forall {Δ : seq ty}, subst Γ Δ -> pnat (Ctx Δ) (Ctx Γ) :=
  match Γ as Γ return forall {Δ : seq ty}, subst Γ Δ -> pnat (Ctx Δ) (Ctx Γ) with
  | [::] => fun _ _ => !
  | A :: Γ => fun Δ τ => ⟨ Subst (fun B i => τ B (inr i)), Tmn (τ A (inl erefl)) ⟩
  end.

Lemma Pos_ren Γ Δ (ξ : ren Γ Δ) :
  forall A (i : pos Γ A), Pos (ξ A i) = Ren ξ ∘ Pos i.
Proof.
  elim: Γ Δ ξ => //= B Γ ih Δ ξ A [E|i]/=.
  - destruct E. exact: pnat_eq.
  - rewrite (ih _ (fun _ i => ξ _ (inr i))) /=. exact: pnat_eq.
Qed.

Lemma Ren_comp Γ Δ Θ (ξ₁ : ren Γ Δ) (ξ₂ : ren Δ Θ) :
  Ren (fun A i => ξ₂ A (ξ₁ A i)) = Ren ξ₂ ∘ Ren ξ₁.
Proof.
  elim: Γ Δ Θ ξ₁ ξ₂ => [|A Γ ih] Δ Θ ξ₁ ξ₂.
  - rewrite-> unit_eta. exact: unit_eta.
  - apply: pnat_eq; fext=> /= x ρ. f_equal. by rewrite ih.
      by rewrite Pos_ren.
Qed.

(** ** Ctx preserves finite-products *)

(*Lemma Ctx_empty : Ctx [::] = 1.
Proof. reflexivity. Qed.*)

Fixpoint Ctx_prod_pi1 Γ Δ : pnat (Ctx (ctx_prod Γ Δ)) (Ctx Γ) :=
  match Γ with
  | [::] => !
  | A :: Γ => first _ (Ctx_prod_pi1 _ _)
  end.

Fixpoint Ctx_prod_pi2 Γ Δ : pnat (Ctx (ctx_prod Γ Δ)) (Ctx Δ) :=
  match Γ with
  | [::] => cid
  | A :: Γ => π₁ ∘ Ctx_prod_pi2 _ _
  end.

Fixpoint Ctx_prod_pair Γ Δ Θ :
  forall (f : pnat Γ (Ctx Δ)) (g : pnat Γ (Ctx Θ)), pnat Γ (Ctx (ctx_prod Δ Θ)) :=
  match Δ with
  | [::] => fun _ g => g
  | A :: Γ => fun f g => ⟨ Ctx_prod_pair (f ∘ π₁) g, f ∘ π₂ ⟩
  end.

Lemma bangP {D : cartcat} (a : D) (f g : D a 1) : f = g.
Proof. rewrite-> unit_eta. exact: unit_eta. Qed.

Lemma pairP {D : cartcat} (a b c : D) (f g : D a (b × c)) :
  f ∘ π₁ = g ∘ π₁ -> f ∘ π₂ = g ∘ π₂ -> f = g.
Proof.
  move=> e1 e2. rewrite -[f]pair_eta e1 e2. exact: pair_eta.
Qed.

(* broken rewrites *)
Lemma Ctx_prod_pi1_beta Γ Δ Θ (f : pnat Γ (Ctx Δ)) (g : pnat Γ (Ctx Θ)) :
  Ctx_prod_pair f g ∘ Ctx_prod_pi1 _ _ = f.
Proof.
  elim: Δ Γ Θ f g => [|A Δ ih] Γ Θ /= f g. exact: bangP.
  apply: pairP.
  - erewrite <-ih. instantiate (2 := g).
    unfold first. csimpl.
    transitivity (⟨Ctx_prod_pair (f ∘ π₁) g, f ∘ π₂⟩ ∘ (π₁ ∘ Ctx_prod_pi1 Δ Θ)).
    f_equal. apply pair_pi1. apply pair_pi1_comp.
  - unfold first. csimpl.
    transitivity (⟨Ctx_prod_pair (f ∘ π₁) g, f ∘ π₂⟩ ∘ π₂).
    f_equal. apply pair_pi2. apply pair_pi2.
Qed.

Lemma Ctx_prod_pi2_beta Γ Δ Θ (f : pnat Γ (Ctx Δ)) (g : pnat Γ (Ctx Θ)) :
  Ctx_prod_pair f g ∘ Ctx_prod_pi2 _ _ = g.
Proof.
  elim: Δ Γ Θ f g => [|A Δ ih] Γ Θ /= f g. by csimpl.
  transitivity (Ctx_prod_pair (f ∘ π₁) g ∘ Ctx_prod_pi2 Δ Θ).
  apply pair_pi1_comp. apply ih.
Qed.

Lemma Ctx_prod_eta Γ Δ Θ (f : pnat Γ (Ctx (ctx_prod Δ Θ))) :
  Ctx_prod_pair (f ∘ Ctx_prod_pi1 _ _) (f ∘ Ctx_prod_pi2 _ _) = f.
Proof.
  elim: Δ Γ Θ f => [|A Δ ih] Γ Θ /= f. by csimpl.
  apply: pairP.
  - csimpl. generalize (ih _ _ (f ∘ π₁)) => H.
    transitivity (Ctx_prod_pair (f ∘ π₁ ∘ Ctx_prod_pi1 Δ Θ) (f ∘ π₁ ∘ Ctx_prod_pi2 Δ Θ)); last first.
    exact: H. csimpl. f_equal. f_equal. unfold first. apply pair_pi1.
  - csimpl. f_equal. apply pair_pi2.
Qed.

Definition Ctx_prod Γ Δ : pnat (Ctx (ctx_prod Γ Δ)) (Ctx Γ × Ctx Δ) :=
  ⟨ Ctx_prod_pi1 _ _, Ctx_prod_pi2 _ _ ⟩.

Definition Ctx_prod_inv Γ Δ : pnat (Ctx Γ × Ctx Δ) (Ctx (ctx_prod Γ Δ)) :=
  Ctx_prod_pair π₁ π₂.

Lemma Ctx_prod_sect Γ Δ :
  Ctx_prod Γ Δ ∘ Ctx_prod_inv Γ Δ = cid.
Proof.
  rewrite/Ctx_prod/Ctx_prod_inv. rewrite -[LHS in LHS = _]Ctx_prod_eta.
  csimpl.
  transitivity (Ctx_prod_pair
                  (⟨Ctx_prod_pi1 Γ Δ, Ctx_prod_pi2 Γ Δ⟩ ∘ π₁)
                  (⟨Ctx_prod_pi1 Γ Δ, Ctx_prod_pi2 Γ Δ⟩ ∘ π₂)).
  f_equal. f_equal. apply Ctx_prod_pi1_beta. f_equal. apply Ctx_prod_pi2_beta.
  csimpl. transitivity (Ctx_prod_pair (cid ∘ Ctx_prod_pi1 Γ Δ) (cid ∘ Ctx_prod_pi2 Γ Δ)).
    by csimpl. apply Ctx_prod_eta.
Qed.

Lemma Ctx_prod_retr Γ Δ :
  Ctx_prod_inv Γ Δ ∘ Ctx_prod Γ Δ = cid.
Proof.
  rewrite/Ctx_prod/Ctx_prod_inv. apply: pairP; csimpl.
  - transitivity (Ctx_prod_pair π₁ π₂ ∘ Ctx_prod_pi1 Γ Δ). f_equal. apply pair_pi1.
    apply Ctx_prod_pi1_beta.
  - transitivity (Ctx_prod_pair π₁ π₂ ∘ Ctx_prod_pi2 Γ Δ). f_equal.
    apply pair_pi2. apply Ctx_prod_pi2_beta.
Qed.

(** ** Action of Ren on projections *)

Definition ctx_inr' Γ Δ : ren Δ (ctx_prod Γ Δ).
Proof.
  induction Γ. cbn. exact (fun _ => id).
  cbn. exact (fun A i => inr (IHΓ A i)).
Defined.

Lemma ctx_inr_inr' Γ Δ : @ctx_inr Γ Δ = @ctx_inr' Γ Δ.
Proof.
  elim: Γ Δ => //=A Γ ih1 Δ. fext.
  move=> /= B i. congr inr. by rewrite ih1.
Qed.

Lemma Ren_inr Γ Δ :
  Ren (@ctx_inr Δ Γ) = Ctx_prod_pi2 Δ Γ.
Proof.
  rewrite ctx_inr_inr'.

  elim: Γ Δ => /=. intros. exact: bangP. move=> A Γ ih Δ /=. elim: Δ => /=.
  - move: (ih [:: A]) => /=. csimpl. move->. by csimpl.
  - move=> B Δ ih'. apply: pairP. csimpl.
    rewrite (@Ren_comp _ _ _ _ shift).
    rewrite -ih'. csimpl. f_equal.
    move: (ih (ctx_prod Δ [:: A])) => /=.
    Set Printing All.

    move: (ih [:: B]) => /=. csimpl.
    rewrite 
    apply: ih.
    unfold ctx_inr => /=.
  move=>/=. unfold ctx_inr. cbn.
  unfold ctx_inr in ih. rewrite ih. rewrite (@Ren_comp _ _ _ _ shift).

Lemma Ren_inr Γ :
  forall Δ, (Ren (@ctx_inr Γ Δ) = Ctx_prod_pi2 Γ Δ) /\ Ren (@ctx_inl Γ Δ) = Ctx_prod_pi1 Γ Δ.
Proof.
  move=> Δ. elim: Δ Γ => [|A Δ ih] Γ /=.
  - split. exact: bangP. induction Γ => //=. unfold Ren Ctx_prod_pi1.
  elim: Γ => [|A Γ ih] Δ /=.
  - unfold ctx_inr. cbn. rewrite-> pair_pi1.
  move=> Δ. elim: Δ Γ => [|A Δ ih] Γ /=. exact: bangP.
  apply: pairP. csimpl.
  rewrite (@Ren_comp _ _ _ shift).
  rewrite (@Pos_ren _ _ _ shift). (@ctx_inr _ _)).
  elim: Γ => [|A Γ ih] Δ /=.
  - unfold ctx_inr. cbn. rewrite-> pair_pi1.
    by rewrite-> unit_eta.
  apply: pnat_eq; fext=> /= x ρ.
  rewrite (@Pos_ren _ _ (@ctx_inr _ _)).
  cbn. rewrite-> ih.
  rewrite (@Ren_comp _ _ _ shift). rewrite ih.


Lemma Ren_inl Γ :
  forall Δ, Ren (@ctx_inl Γ Δ) = Ctx_prod Γ Δ ∘ π₂.
Proof.
  elim: Γ => [|A Γ ih] Δ /=. by rewrite-> unit_eta.
  csimpl. unfold first. f_equal. rewrite (@Ren_comp _ _ _ _ shift).
  rewrite ih.
  rewrite (@Ren_comp _ _ _ shift shift) ih.
  
  Set Printing All.
  rewrite ih.

Lemma Ren_id Γ :
  Ren (fun A (i : pos Γ A) => i) = cid.
Proof.
  elim: Γ => /=. by rewrite-> unit_eta. move=> A Γ ih.
  rewrite <-pair_eta_id.

Lemma TmnTm_ren Γ :
  (forall A (s : tmn Γ A),
      forall Δ (ξ : ren Γ Δ), Tmn (tmn_ren s ξ) = Ren ξ ∘ Tmn s) /\
  (forall A (s : tm Γ A),
      forall Δ (ξ : ren Γ Δ), Tm (tm_ren s ξ) = Ren ξ ∘ Tm s).
Proof.
  move: Γ. apply: t_ind => //.
  - move=> Γ A i Δ ξ /=. exact: Pos_ren.
  - move=> Γ A B s ihs t iht Δ ξ. apply: pnat_eq; fext=> x ρ /=.
      by rewrite ihs iht.
  - move=> Γ σ B s ih v Δ ξ. apply: pnat_eq; fext=> x ρ /=. by rewrite ih.
  - move=> Γ A B s ih Δ ξ. apply: pnat_eq; fext=> x ρ.
    apply: sig_eq => /=. fext=> b h /= t. rewrite ih. cbn.
    f_equal. f_equal. rewrite -natP. rewrite (@Ren_comp _ _ _ ξ shift) /=. (* lemma *) admit.
  - move=> Γ c B b ih Δ ξ. apply: pnat_eq; fext=> x ρ.
    move=> v /=. by rewrite ih.
Qed.


End Interpretation.

(** ** Kripke LR interpretation *)

Section RelationalInterpretation.

Variable (C D : cat) (d1 : presheaf C) (d2 : presheaf D).

Variable (Def : forall x y, d1 x -> d2 y -> Prop).
Hypothesis (DefP : forall (x1 y1 : C) (x2 y2 : D) (f : hom y1 x1) (g : hom y2 x2),
             forall (a : d1 x1) (b : d2 x2), Def a b -> Def (map f a) (map g b)).

Fixpoint DTy (A : ty) : forall x y, @Ty C d1 A x -> @Ty D d2 A y -> Prop :=
  match A as A return forall x y, @Ty C d1 A x -> @Ty D d2 A y -> Prop with
  | DefTy => Def
  | Arrow A B =>
    fun x y F G =>
      forall x' y' (f : C x' x) (g : D y' y) (a : Ty d1 A x') (b : Ty d2 A y'),
        @DTy A _ _ a b -> @DTy B _ _ (F.1 x' f a) (G.1 y' g b)
  | Prod σ B =>
    fun x y F G => forall i : ext σ, DTy (F i) (G i)
  end.

Lemma DTyP (A : ty) :
  forall (x1 y1 : C) (x2 y2 : D) (f : hom y1 x1) (g : hom y2 x2),
  forall (a : Ty d1 A x1) (b : Ty d2 A x2), DTy a b -> DTy (map f a) (map g b).
Proof.
  elim: A => //=.
  - move=> A ih1 B ih2 x1 y1 x2 y2 f1 g1 [F HF] [G HG] /= HA z1 z2 f2 g2 a b dab.
    exact: HA.
  - move=> σ B ih x1 y1 x2 y2 f g a b HA i. apply: ih. exact: HA.
Qed.

Fixpoint DCtx (Γ : seq ty) : forall x y, Ctx d1 Γ x -> Ctx d2 Γ y -> Prop :=
  match Γ as Γ return forall x y, Ctx d1 Γ x -> Ctx d2 Γ y -> Prop with
  | [::] => fun _ _ _ _ => True
  | A :: Δ => fun x y (ρ₁ : Ctx d1 Δ x * Ty d1 A x) (ρ₂ : Ctx d2 Δ y * Ty d2 A y) =>
                DCtx (fst ρ₁) (fst ρ₂) /\ DTy (snd ρ₁) (snd ρ₂)
  end.

Lemma DCtxP (Γ : seq ty) :
  forall (x1 y1 : C) (x2 y2 : D) (f : hom y1 x1) (g : hom y2 x2),
  forall (ρ₁ : Ctx d1 Γ x1) (ρ₂ : Ctx d2 Γ x2), DCtx ρ₁ ρ₂ -> DCtx (map f ρ₁) (map g ρ₂).
Proof.
  elim: Γ => // A Γ ih x1 y1 x2 y2 f g /= [ρ₁ a] [ρ₂ b] /= [Hρ HA].
  split. exact: ih. exact: DTyP.
Qed.

Lemma DPos (Γ : seq ty) (A : ty) (i : pos Γ A) :
  forall x y (ρ₁ : Ctx d1 Γ x) (ρ₂ : Ctx d2 Γ y),
    DCtx ρ₁ ρ₂ -> DTy (Pos d1 i x ρ₁) (Pos d2 i y ρ₂).
Proof.
  induction Γ; cbn. destruct i.
  destruct 0; cbn. destruct 1. cbn. cbn in i.
  destruct i. destruct e. assumption.
  apply IHΓ. assumption.
Defined.

Lemma DTmnDTm (Γ : seq ty) :
  (forall (A : ty) (s : tmn Γ A),
      forall x y (ρ₁ : Ctx d1 Γ x) (ρ₂ : Ctx d2 Γ y),
        DCtx ρ₁ ρ₂ -> DTy (Tmn d1 s x ρ₁) (Tmn d2 s y ρ₂)) /\
  (forall (A : ty) (s : tm Γ A),
      forall x y (ρ₁ : Ctx d1 Γ x) (ρ₂ : Ctx d2 Γ y),
        DCtx ρ₁ ρ₂ -> DTy (Tm d1 s x ρ₁) (Tm d2 s y ρ₂)).
Proof.
  move: Γ. apply: t_ind.
  - move=> Γ A i x y ρ₁ ρ₂ Hρ /=. exact: DPos.
  - move=> Γ A B s ihs t iht x y ρ₁ ρ₂ Hρ.
    specialize (ihs x y ρ₁ ρ₂ Hρ).
    specialize (iht x y ρ₁ ρ₂ Hρ).
    cbn in ihs.
    specialize (ihs x y cid cid _ _ iht) => {iht}. exact: ihs.
  - move=> Γ σ B s ihs v x y ρ₁ ρ₂ Hρ. exact: ihs.
  - move=> Γ A B s ihs x y ρ₁ ρ₂ Hρ.
    cbn. move=> x' y' f g a b dab. apply: ihs.
    cbn. split=> //. exact: DCtxP.
  - move=> Γ c B b ih x y ρ₁ ρ₂ Hρ /= i; exact: ih.
  - move=> Γ s ih x y ρ₁ ρ₂ Hρ. exact: ih.
Qed.

Lemma DTmn (Γ : seq ty) (A : ty) (s : tmn Γ A) x y (ρ₁ : Ctx d1 Γ x) (ρ₂ : Ctx d2 Γ y) :
  DCtx ρ₁ ρ₂ -> DTy (Tmn d1 s x ρ₁) (Tmn d2 s y ρ₂).
Proof. case: (DTmnDTm Γ); eauto. Qed.

Lemma DTm (Γ : seq ty) (A : ty) (s : tm Γ A) x y (ρ₁ : Ctx d1 Γ x) (ρ₂ : Ctx d2 Γ y) :
  DCtx ρ₁ ρ₂ -> DTy (Tm d1 s x ρ₁) (Tm d2 s y ρ₂).
Proof. case: (DTmnDTm Γ); eauto. Qed.


(*** Standard Interpretation ***)

Section Interpretation.

Variable (def : Type).

(* standard interpretation *)

Fixpoint Ty (A : ty) : Type :=
  match A with
  | DefTy => def
  | Arrow A B => Ty A -> Ty B
  | Prod σ B => ext σ -> Ty B
  end.

Fixpoint Ctx (Γ : seq ty) : Type :=
  match Γ with
  | [::] => unit
  | A :: Δ => Ctx Δ * Ty A
  end.

Definition Pos (Γ : seq ty) (A : ty) (p : pos Γ A) (γ : Ctx Γ) : Ty A.
  induction Γ. destruct p. cbn in p. destruct p as [E|p].
  - cbn in γ. exact (match E with erefl => snd γ end).
  - apply IHΓ. exact p. exact (fst γ).
Defined.

Fixpoint Tmn (Γ : seq ty) (A : ty) (s : tmn Γ A) : Ctx Γ -> Ty A :=
  match s with
  | var _ i => Pos i
  | app _ _ s t => fun ρ => Tmn s ρ (Tm t ρ)
  | proj _ _ s v => fun ρ => Tmn s ρ v
  end
with Tm (Γ : seq ty) (A : ty) (s : tm Γ A) : Ctx Γ -> Ty A :=
  match s with
  | lam A B s => fun (ρ : Ctx Γ) (x : Ty A) => @Tm (A :: Γ) B s (ρ, x)
  | abs c B b => fun (ρ : Ctx Γ) (v : ext c) => Tm (b v) ρ
  | neu s => Tmn s
  end.

(* predicate interpretation *)

Variable Def : def -> Type.

Fixpoint DTy (A : ty) : Ty A -> Type :=
  match A with
  | DefTy => Def
  | Arrow A B => fun f : Ty A -> Ty B => forall x : Ty A, DTy x -> DTy (f x)
  | Prod σ B => fun f : ext σ -> Ty B => forall x : ext σ, DTy (f x)
  end.

Fixpoint DCtx (Γ : seq ty) : Ctx Γ -> Type :=
  match Γ with
  | [::] => fun _ => unit
  | A :: Δ => fun (ρ : Ctx Δ * Ty A) => (DCtx (fst ρ) * DTy (snd ρ))%type
  end.

Lemma DPos (Γ : seq ty) (A : ty) (i : pos Γ A) :
  forall (ρ : Ctx Γ), DCtx ρ -> DTy (Pos i ρ).
Proof.
  induction Γ; cbn. destruct i.
  destruct 0; cbn. destruct 1. cbn. cbn in i.
  destruct i. destruct e. assumption.
  apply IHΓ. assumption.
Defined.

Fixpoint DTmn (Γ : seq ty) (A : ty) (s : tmn Γ A) :
  forall ρ : Ctx Γ, DCtx ρ -> DTy (Tmn s ρ) :=
  match s with
  | var _ i => DPos i
  | app A B s t => fun ρ ρ' => @DTmn Γ (Arrow A B) s ρ ρ' (Tm t ρ) (@DTm Γ A t ρ ρ')
  | proj σ B s v => fun ρ ρ' => @DTmn Γ (Prod σ B) s ρ ρ' v
  end
with DTm (Γ : seq ty) (A : ty) (s : tm Γ A) :
  forall ρ : Ctx Γ, DCtx ρ -> DTy (Tm s ρ) :=
  match s with
  | lam A B s => fun ρ ρ' (x : Ty A) (x' : DTy x) =>
                   @DTm (A :: Γ) B s (ρ, x) (ρ', x')
  | abs σ B b => fun ρ ρ' (x : ext σ) => @DTm Γ B (b x) ρ ρ'
  | neu s => DTmn s
  end.

End Interpretation.

Definition lambda : seq ty :=
  [:: Arrow DefTy (Arrow DefTy DefTy); Arrow (Arrow DefTy DefTy) DefTy ].

Variable (T : Type) (P : T -> Type).
Compute @Ctx T lambda.
Variable (app : T -> T -> T).
Variable (lam : (T -> T) -> T).
Definition ρ : @Ctx T lambda := ((tt, lam), app).
Compute @DCtx T P lambda ρ.
Eval hnf in forall s, @DTy T P DefTy (@Tmn T lambda DefTy s ρ).