Require Import base ord.

(** ** The Chu construction over Prop *)

Reserved Notation "x ^+" (at level 3, left associativity, format "x ^+").
Reserved Notation "x ^-" (at level 3, left associativity, format "x ^-").
Reserved Notation "x ^⊥" (at level 3, left associativity, format "x ^⊥").

Section Chu.
  Variable (d : Prop).
  Implicit Types (x y z : Prop).

  Record chu := mk_chu { chuL : Prop; chuR : Prop; _ : chuL -> chuR -> d }.
  Arguments mk_chu l r p : rename, clear implicits.
  Implicit Types (p q r : chu).

  Notation "p ^+" := (chuL p).
  Notation "p ^-" := (chuR p).

  Lemma chuP p : p^+ -> p^- -> d.
  Proof. by case: p. Qed.

  (*Program Definition dual p := mk_chu (chuR p) (chuL p) _.
  Next Obligation. rewrite starC. exact: chuP. Qed.

  Notation "p ^⊥" := (dual p).*)

  Definition chu_le p q := (chuL p ≤ chuL q) /\ (chuR q ≤ chuR p).

  Lemma chu_le_refl : reflexive chu_le.
  Proof. by split. Qed.

  Lemma chu_le_trans : transitive chu_le.
  Proof. move=> p q r [h1 h2] [h3 h4]. split. by rewrite h1. by rewrite h4. Qed.

  Canonical chu_proMixin := ProMixin chu_le_refl chu_le_trans.
  Canonical chu_proType  := Eval hnf in ProType chu chu_proMixin.

  Lemma chu_eq p q : chuL p = chuL q -> chuR p = chuR q -> p = q.
  Proof.
    move: p q => [l1 r1 h1] [l2 r2 h2] /= e1 e2. destruct e1, e2.
    f_equal. exact: pi.
  Qed.

  Lemma chu_le_eq p q : p ≤ q -> q ≤ p -> p = q.
  Proof.
    move=> [l1 r1] [l2 r2]. apply: chu_eq; exact: le_eq.
  Qed.

  Canonical chu_ordMixin := OrdMixin chu_le_eq.
  Canonical chu_ordType  := Eval hnf in OrdType chu chu_ordMixin.

  (*Lemma le_dual p q :
    (p ≤ q) <-> (q^⊥ ≤ p^⊥).
  Proof. split=> -[]; by split. Qed.*)

  Program Definition chu_inf T (F : T -> chu) := mk_chu (inf (chuL \o F)) (sup (chuR \o F)) _.
  Next Obligation.
    rewrite prop_exE in H0. case: H0 => y /=. apply: chuP. exact: H.
  Qed.

  Lemma chu_infP T (F : T -> chu) : is_glb F (chu_inf F).
  Proof.
    move=> x /=. split. move=> [/=h1 h2] i. split. rewrite h1. exact: allE (i) _.
    rewrite -h2. exact: exI (i) _. move=> h. split=> /=. apply: allI => i /=.
    by case: (h i). apply: exE => i /=. by case: (h i).
  Qed.

  Canonical chu_clatMixin := CLatMixin chu_infP.
  Canonical chu_clat := Eval hnf in CLat chu chu_clatMixin.

  Program Definition chu_dual p := mk_chu (chuR p) (chuL p) _.
  Next Obligation. exact: chuP H. Qed.
  Notation "p ^⊥" := (chu_dual p).

  Lemma chu_dual_pos p : p^⊥^+ = p^-.
  Proof. by []. Qed.

  Lemma chu_dual_neg p : p^⊥^- = p^+.
  Proof. by []. Qed.

  Lemma chu_topEp : ⊤^+ = ⊤.
  Proof. apply: top_eq => /=; by focus. Qed.

  Lemma chu_topEn : ⊤^- = ⊥.
  Proof. apply: bot_eq => /=; by focus. Qed.

  Lemma chu_botEp : ⊥^+ = ⊥.
  Proof. apply: bot_eq => /=. rewrite prop_leE prop_allE prop_botE /= => H. by apply: (H (mk_chu False True _)). Qed.
End Chu.


(*

  Lemma chu_botEn : ⊥^- = ⊤.
  Proof. by rewrite -dual_top chu_dual_neg chu_topEp. Qed.

  Lemma chu_meetEp (x y : chu) :
    (x ∧ y)^+ = (x^+ ∧ y^+).
  Proof.
    apply: le_eq. focus. exact: allE true _. exact: allE false _.
    cbn. focus=> -[] /=. exact: meetEl. exact: meetEr.
  Qed.

  Lemma chu_meetEn (x y : chu) :
    (x ∧ y)^- = (x^- ∨ y^-).
  Proof.
    apply: le_eq => /=; focus. case=>/=. exact: joinIl. exact: joinIr.
    exact: exI true _. exact: exI false _.
  Qed.

  Lemma chu_joinEp (x y : chu) :
    (x ∨ y)^+ = (x^+ ∨ y^+).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_pos chu_meetEn !chu_dual_neg.
  Qed.

  Lemma chu_joinEn (x y : chu) :
    (x ∨ y)^- = (x^- ∧ y^-).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_neg chu_meetEp !chu_dual_pos.
  Qed.

  Lemma chu_infEp I (F : I -> chu) :
    (inf F)^+ = ∀ i, (F i)^+.
  Proof. by []. Qed.

  Lemma chu_infEn I (F : I -> chu) :
    (inf F)^- = ∃ i, (F i)^-.
  Proof. by []. Qed.

  Lemma dual_supE {C : cll} T (F : T -> C) :
    sup F = (∀ i, (F i)^⊥)^⊥.
  Proof.
    rewrite dual_inf. f_equal; fext=> i. by rewrite dual_involution.
  Qed.

  Lemma chu_supEp T (F : T -> chu) :
    (sup F)^+ = (∃ i, (F i)^+).
  Proof.
    rewrite dual_supE chu_dual_pos chu_infEn.
      by setoid_rewrite chu_dual_neg.
  Qed.

  Lemma chu_supEn T (F : T -> chu) :
    (sup F)^- = (∀ i, (F i)^-).
  Proof.
    rewrite dual_supE chu_dual_neg chu_infEp.
      f_equal; fext=> i. by rewrite chu_dual_pos.
  Qed.

  Lemma chu_starEp p q :
    (p ⊗ q)^+ = (p^+ ⊗ q^+).
  Proof. by []. Qed.

  Lemma chu_starEn p q :
    (p ⊗ q)^- = ((p^+ -∘ q^-) ∧ (q^+ -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEp p q :
    (p -∘ q)^+ = ((p^+ -∘ q^+) ∧ (q^- -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEn p q :
    (p -∘ q)^- = (p^+ ⊗ q^-).
  Proof. by []. Qed.

  Lemma chu_IEp :
    I^+ = I.
  Proof. by []. Qed.

  Lemma chu_IEn :
    I^- = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEp : O^+ = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEn : O^- = I.
  Proof. by []. Qed.

  Lemma chu_parEp p q :
    (p ⊕ q)^+ = ((p^- -∘ q^+) ∧ (q^- -∘ p^+)).
  Proof.
    by rewrite/par chu_dual_pos chu_starEn !chu_dual_pos !chu_dual_neg.
  Qed.

  Lemma chu_parEn p q :
    (p ⊕ q)^- = (p^- ⊗ q^-).
  Proof.
    by rewrite/par chu_dual_neg chu_starEp !chu_dual_pos.
  Qed.

  Program Definition chu_embed x : chu :=
    mk_chu x (x -∘ ⊥) _.
  Next Obligation. exact: starxW. Qed.

  Program Definition chu_star p q :=
    mk_chu (p^+ /\ q^+) ((p^+ -> q^-) /\ (q^+ -> p^-)) _.
  Next Obligation.
    apply: chuP H _. exact: H1.
  Qed.

  Definition chu_wand p q := chu_dual (chu_star p (chu_dual q)).

  Program Definition chu_I := mk_chu True d _.

  Lemma chu_starC p q : chu_star p q ≤ chu_star q p.
  Proof. split=> /=. by rewrite starC. by rewrite meetC. Qed.

  Lemma chu_star1x p : chu_star chu_I p = p.
  Proof.
    apply: le_eq; split=> /=.
    - by rewrite star1x.
    - rewrite wand1x. focus=> //. apply/star_adj.
      rewrite starC. exact: chuP.
    - by rewrite star1x.
    - rewrite wand1x. exact: meetEl.
  Qed.

  Lemma chu_starA p q r :
    chu_star (chu_star p q) r ≤ chu_star p (chu_star q r).
  Proof.
    split=> /=. by rewrite starA. apply: meetI.
    apply: meetEl. rewrite wandxM. apply: meetEl. by rewrite wandxW.
    rewrite !wandxM. apply: meetI. apply: meetEl. apply: meetEr.
    by rewrite !wandxW starC. apply: meetEr. by rewrite wandxW starC.
  Qed.

  Lemma chu_star_adj p q r :
    (chu_star p q ≤ r) <-> (p ≤ chu_wand q r).
  Proof.
    split=> -[/= h1 h2]; split=> /=.
    - apply: meetI. exact/star_adj. apply/star_adj. rewrite h2.
      rewrite starxM. apply: meetEl. by rewrite starxW.
    - rewrite h2. rewrite starxM. apply: meetEr. by rewrite starxW.
    - rewrite h1. rewrite starMx. apply: meetEl. by rewrite starWx.
    - apply: meetI; apply/star_adj; last first. by rewrite starC.
      rewrite h1 starxM. apply: meetEr. by rewrite starxW.
  Qed.

  Canonical chu_quantaleMixin := QuantaleMixin chu_star_adj chu_starC chu_starA chu_star1x.
  Canonical chu_quantale := Eval hnf in Quantale chu chu_quantaleMixin.

  Definition chu_O := chu_dual I.

  Lemma chu_dualP p :
    (p -∘ chu_O) = chu_dual p.
  Proof.
    apply: chu_eq => /=; last first. by rewrite starx1.
    rewrite wand1x. apply: le_eq. exact: meetEr.
    focus. apply/star_adj. rewrite starC. exact: chuP.
  Qed.

  Lemma chu_classical p :
    ((p -∘ chu_O) -∘ chu_O) ≤ p.
  Proof.
    rewrite !chu_dualP. case: p => /=l r o. by split.
  Qed.

  Canonical chu_cllMixin := CLLMixin chu_classical.
  Canonical chu_cll := Eval hnf in CLL chu chu_cllMixin.

  Lemma chu_dualE p : p^⊥ = chu_dual p.
  Proof. exact: chu_dualP. Qed.

  Lemma chu_dual_pos p : p^⊥^+ = p^-.
  Proof. by rewrite chu_dualE. Qed.

  Lemma chu_dual_neg p : p^⊥^- = p^+.
  Proof. by rewrite chu_dualE. Qed.

  Lemma chu_topEp : ⊤^+ = ⊤.
  Proof. apply: top_eq => /=; by focus. Qed.

  Lemma chu_topEn : ⊤^- = ⊥.
  Proof. apply: bot_eq => /=; by focus. Qed.

  Lemma chu_botEp : ⊥^+ = ⊥.
  Proof. by rewrite -dual_top chu_dual_pos chu_topEn. Qed.

  Lemma chu_botEn : ⊥^- = ⊤.
  Proof. by rewrite -dual_top chu_dual_neg chu_topEp. Qed.

  Lemma chu_meetEp (x y : chu) :
    (x ∧ y)^+ = (x^+ ∧ y^+).
  Proof.
    apply: le_eq. focus. exact: allE true _. exact: allE false _.
    cbn. focus=> -[] /=. exact: meetEl. exact: meetEr.
  Qed.

  Lemma chu_meetEn (x y : chu) :
    (x ∧ y)^- = (x^- ∨ y^-).
  Proof.
    apply: le_eq => /=; focus. case=>/=. exact: joinIl. exact: joinIr.
    exact: exI true _. exact: exI false _.
  Qed.

  Lemma chu_joinEp (x y : chu) :
    (x ∨ y)^+ = (x^+ ∨ y^+).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_pos chu_meetEn !chu_dual_neg.
  Qed.

  Lemma chu_joinEn (x y : chu) :
    (x ∨ y)^- = (x^- ∧ y^-).
  Proof.
    rewrite -{1}[x]dual_involution -{1}[y]dual_involution -dual_meet.
      by rewrite chu_dual_neg chu_meetEp !chu_dual_pos.
  Qed.

  Lemma chu_infEp I (F : I -> chu) :
    (inf F)^+ = ∀ i, (F i)^+.
  Proof. by []. Qed.

  Lemma chu_infEn I (F : I -> chu) :
    (inf F)^- = ∃ i, (F i)^-.
  Proof. by []. Qed.

  Lemma dual_supE {C : cll} T (F : T -> C) :
    sup F = (∀ i, (F i)^⊥)^⊥.
  Proof.
    rewrite dual_inf. f_equal; fext=> i. by rewrite dual_involution.
  Qed.

  Lemma chu_supEp T (F : T -> chu) :
    (sup F)^+ = (∃ i, (F i)^+).
  Proof.
    rewrite dual_supE chu_dual_pos chu_infEn.
      by setoid_rewrite chu_dual_neg.
  Qed.

  Lemma chu_supEn T (F : T -> chu) :
    (sup F)^- = (∀ i, (F i)^-).
  Proof.
    rewrite dual_supE chu_dual_neg chu_infEp.
      f_equal; fext=> i. by rewrite chu_dual_pos.
  Qed.

  Lemma chu_starEp p q :
    (p ⊗ q)^+ = (p^+ ⊗ q^+).
  Proof. by []. Qed.

  Lemma chu_starEn p q :
    (p ⊗ q)^- = ((p^+ -∘ q^-) ∧ (q^+ -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEp p q :
    (p -∘ q)^+ = ((p^+ -∘ q^+) ∧ (q^- -∘ p^-)).
  Proof. by []. Qed.

  Lemma chu_wandEn p q :
    (p -∘ q)^- = (p^+ ⊗ q^-).
  Proof. by []. Qed.

  Lemma chu_IEp :
    I^+ = I.
  Proof. by []. Qed.

  Lemma chu_IEn :
    I^- = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEp : O^+ = ⊥.
  Proof. by []. Qed.

  Lemma chu_OEn : O^- = I.
  Proof. by []. Qed.

  Lemma chu_parEp p q :
    (p ⊕ q)^+ = ((p^- -∘ q^+) ∧ (q^- -∘ p^+)).
  Proof.
    by rewrite/par chu_dual_pos chu_starEn !chu_dual_pos !chu_dual_neg.
  Qed.

  Lemma chu_parEn p q :
    (p ⊕ q)^- = (p^- ⊗ q^-).
  Proof.
    by rewrite/par chu_dual_neg chu_starEp !chu_dual_pos.
  Qed.

  Program Definition chu_embed x : chu :=
    mk_chu x (x -∘ ⊥) _.
  Next Obligation. exact: starxW. Qed.

  Lemma chu_box_def p :
    box p = chu_embed (box p^+).
  Proof.
    apply: le_eq.
    - rewrite{1}/box. focus=> /=i [[[/= aff1 aff2] [/= dup1 dup2]] [le_i_p le_p_i]]. split=> /=.
      + apply: exIc (i^+) _ le_refl. split. split. exact: aff1. exact: dup1. exact: le_i_p.
      + rewrite -dup2 meetxx. apply/star_adj. rewrite aff2.
        suff: i^+ ≤ box p^+. move->. exact: starWx.
        exact: exIc (i^+) _ le_refl.
    - apply: exIc (chu_embed (box p^+)) _ le_refl. split; [split|].
      + split=> /=. exact: is_persistent_is_affine. apply/star_adj.
          by rewrite starBx.
      + split=> /=. by rewrite -dup. rewrite meetxx. by rewrite wandxW -dup.
      + split=> /=. exact: kernel_dec. apply/star_adj. rewrite kernel_dec starC.
        exact: chuP.
  Qed.
  
  Lemma chu_boxEp p : (box p)^+ = box (p^+).
  Proof. by rewrite chu_box_def. Qed.

  Lemma chu_boxEn p : (box p)^- = (box (p^+) -∘ ⊥).
  Proof. by rewrite chu_box_def. Qed.

  Definition why p := dual (box (dual p)).

  Lemma chu_whyEp p : (why p)^+ = (box p^- -∘ ⊥).
  Proof. by rewrite/why chu_dual_pos chu_boxEn chu_dual_pos. Qed.

  Lemma chu_whyEn p : (why p)^- = box p^-.
  Proof. by rewrite/why chu_dual_neg chu_boxEp chu_dual_pos. Qed.

  Lemma why_join p q : why (p ∨ q) = (why p ⊕ why q).
  Proof.
    by rewrite{1}/why dual_join box_star dual_star.
  Qed.

  Lemma why_bot : why ⊥ = O.
  Proof. by rewrite/why dual_bot box_top dual_I. Qed.

  Lemma whyE p q :
    O ≤ q -> q ⊕ q ≤ q -> p ≤ q ->
    why p ≤ q.
  Proof.
    move=> h1 h2 h3. apply/dual_leq. rewrite/why dual_involution. apply: boxI.
    - apply/dual_leq. rewrite dual_involution dual_I. exact: h1.
    - apply/dual_leq. rewrite dual_involution. exact: h2.
    - apply/dual_leq. rewrite !dual_involution. exact: h3.
  Qed.

  (*Lemma chu_embed_wand x y :
    (chu_embed x -∘ chu_embed y)^- ≤ (chu_embed (x -∘ y))^-.
  Proof.
    cbn. by rewrite -star_adj [x ⊗ _]starC starA starxW starWx.
  Qed.

  Lemma chu_embed_wand x y :
    (chu_embed x -∘ chu_embed y)^+ = (x -∘ y).
  Proof.
    apply: le_eq => /=.
    - exact: meetEl.
    - focus. by rewrite -!star_adj starC -starA starxW starxW.
  Qed.*)
End Chu.
 *)
